{$S-,G+,R-}
{***********************}
{*    unit Scheduler   *}
{***********************}

unit Schedul;
interface
uses Objects,Process,Dos;
const
  SliceVec = $08;
type
  TickEmulate = (NoTick,RandomTick,FullTick);

  PProcessCollection = ^TProcessCollection;
  TProcessCollection = object(TSortedCollection)
        function Compare(Key1, Key2: Pointer): Integer; virtual;
        function KeyOf(Item: Pointer): Pointer; virtual;
  end;

  PSignal = ^TSignal;
  TSignal = object(TObject)
      Count : integer;
      ProcessCollection: PProcessCollection;

      constructor Init;
      destructor Done; virtual;

      procedure Wait; virtual;
      procedure Send; virtual;
      procedure Notify; virtual;
      function AWaited: Boolean; virtual;
  end;

  PScheduler = ^TScheduler;
  TScheduler = object(TProcess)
{      ProcessCollection : PProcessCollection;}
      ProcessCollection : PCollection;
      SumPr : integer;
      SaveTimer,NewTimer: pointer;
      Divider: byte;
      LockNest: integer;
      SliceDue: boolean;
      ActiveProcess,Parent: PProcess;

      constructor Init(ALimit, ADelta: Integer);
      destructor Done; virtual;

      procedure Insert(Item: Pointer; Pr: integer); virtual;
      procedure Body; virtual;
      procedure Start; virtual;

      procedure Stop; virtual;
      procedure Go; virtual;

      procedure Delete(Item: Pointer); virtual;
      function NextDue: pointer; virtual;
      procedure Slice(Tick: TickEmulate); virtual;
      procedure Lock; virtual;
      procedure UnLock; virtual;
      function Swap(NextProcess: PProcess): PProcess; virtual;
  end;
var
  Scheduler : PScheduler;
implementation
uses Crt;

var
  NopProc : PProcess;
  
          {   TProcessCollection   }

function TProcessCollection.Compare(Key1, Key2: Pointer): Integer;
begin
   if Integer(Key1^) < Integer(Key2^) then Compare := 1
   else if Integer(Key1^) > Integer(Key2^) then Compare := -1
     else Compare := 0;
end;

function TProcessCollection.KeyOf(Item: Pointer): Pointer;
begin
   KeyOf := @PProcess(Item)^.Prioritet;
end;
       { TSignal }
constructor TSignal.Init;
begin
   TObject.Init;
   Count := 0;
   ProcessCollection := New(PProcessCollection, Init(20,10));
end;

destructor TSignal.Done;
begin
   Dispose(ProcessCollection, Done);
end;

procedure TSignal.Wait;
begin
   pushf;
   cli;
   dec(Count);
   if Count < 0 then
   begin
      Scheduler^.Delete(CurrentProcess);
      ProcessCollection^.Insert(CurrentProcess);
   end;
   Scheduler^.Slice(RandomTick);
   popf;
end;

procedure TSignal.Send;
begin
  pushf;
  cli;
  Notify;
  Scheduler^.Slice(RandomTick);
  popf;
end;

procedure TSignal.Notify;
var
  p: PProcess;
begin
  inc(Count);
  if Count <= 0 then begin
     p := ProcessCollection^.At(0);
     ProcessCollection^.Delete(p);
     Scheduler^.Insert(p,p^.Prioritet);
  end;
end;

function TSignal.AWaited: Boolean;
begin
   AWaited := (Count < 0);
end;

                      { TScheduler }

constructor TScheduler.Init(ALimit, ADelta: Integer);      { StartScheduler }
begin
   TProcess.Init(512);
   GetIntVec(SliceVec,SaveTimer);
{   ProcessCollection := New(PProcessCollection, Init(ALimit, ADelta));
   ProcessCollection^.Duplicates := true;}
   ProcessCollection := New(PCollection, Init(ALimit, ADelta));
   Divider := 0;
   SumPr := 0;
   LockNest := 0;
   SliceDue := false;
end;

destructor TScheduler.Done;
begin
   SetIntVec(SliceVec,SaveTimer);
   Dispose(ProcessCollection, Done);
   TProcess.Done;
   Transfer(NopProc,Parent);
end;

procedure TScheduler.Insert(Item: Pointer; Pr: integer);
begin
  pushf;
  {cli;}
  with PProcess(Item)^ do
    if (SumPr div 2) < MaxInt - Prioritet then begin
       Prioritet := 2 * Pr;
       inc(SumPr,Prioritet);
       CurPr := Prioritet;
       ProcessCollection^.Insert(Item);
     end;
  popf;
end;


procedure TScheduler.Delete(Item: Pointer);
begin
  pushf;
 { cli; }
  with PProcess(Item)^ do begin
    dec(SumPr,Prioritet);
  end;
  ProcessCollection^.Delete(Item);
  popf;
end;


procedure TScheduler.Start;
begin
  if ProcessCollection^.Count > 0 then begin
      LockNest := 0;
      Parent := CurrentProcess;
      Transfer(CurrentProcess,@Self);
  end
  else {Abort('There are no processes at Scheduler')} ;
end;

procedure TScheduler.Stop;
begin
  GetIntVec(SliceVec,NewTimer);
  SetIntVec(SliceVec,SaveTimer);
end;

procedure TScheduler.Go;
begin
  SetIntVec(SliceVec,NewTimer);
end;

procedure TScheduler.Body;
begin
  ActiveProcess := NextDue;
  while true do begin
    IOTransfer(CurrentProcess,ActiveProcess,SliceVec);
    asm
      int SliceVec
    end;
    ActiveProcess := NextDue;
    if LockNest > 0 then SliceDue := true;
  end;
end;

function TScheduler.NextDue: pointer;
var
  MaxPr: integer;
  ProcMax: PProcess;

procedure DefMaxPriority(El: PProcess); far;
begin
  with El^ do begin
     inc(CurPr,Prioritet shr Divider);
     if CurPr > MaxPr then begin
        MaxPr := CurPr;
        ProcMax := El;
     end;
  end;
end;

begin
   MaxPr := -MaxInt;
   ProcessCollection^.ForEach(@DefMaxPriority);
   if LockNest = 0 then NextDue := Swap(ProcMax)
   else NextDue := ActiveProcess;
   with ProcMax^ do
     CurPr := CurPr - (SumPr shr Divider);
   Divider := 0;
end;

procedure TScheduler.Slice(Tick: TickEmulate);
begin
  case Tick of
   NoTick     : Divider := 8;
   RandomTick : Divider := 1;
   FullTick   : Divider := 0;
  end;
  SliceDue := false;
  asm
     sti
     int SliceVec
  end;
end;

procedure TScheduler.Lock;
begin
  pushf;
 { cli;}
  if LockNest < MaxInt then inc(LockNest);
  popf;
end;

procedure TScheduler.UnLock;
begin
  pushf;
{  cli;  }
  if LockNest > 0 then dec(LockNest);
  if (LockNest = 0) and SliceDue then Slice(RandomTick);
  popf;
end;

function TScheduler.Swap(NextProcess: PProcess): PProcess;
begin
   Swap := NextProcess;
end;

begin
  NopProc := New(PProcess, Init(16));
end.
